""" Put this in your `__init__.py` file to affect *all* imports inside the folder `__init__.py` resides in. """

from generalimport import generalimport
generalimport("your", "optional", "dependencies")

"""
`generalimport("*")` makes it handle **all** names (If missing of course)

:warning: `generalimport("*")._scope = None` disables the scope
 - Makes it handle missing imports anywhere
 - For example it will override `pandas` internal custom optional dependency handling
"""



import sys

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_bool(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            bool(fakepackage)

    def test_bytes(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            bytes(fakepackage)

    def test_complex(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            complex(fakepackage)

    def test_float(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            float(fakepackage)

    def test_int(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            int(fakepackage)

    def test_iter(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            iter(fakepackage)

    def test_hash(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            hash(fakepackage)



"""
 - When `generalimport` is instantiated it creates a new importer for `sys.meta_path`.
 - This importer will return 'fake' modules for matching names and scope.
 - The scope ensures only your own imports are faked.
 - The fake module will recursively return a FakeModule instance when asked for an attribute.
 - When used in any way (\_\_call\_\_, \_\_add\_\_, \_\_str\_\_ etc) it raises `generalimport.MissingDependencyException`.
 - This exception has the 'skip-exceptions' from `unittest` and `pytest` as bases, which means that tests will automatically be skipped.
"""



import math
import sys
from unittest import skip

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_abs(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            abs(fakepackage)

    def test_add(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage + 3

    def test_ceil(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            math.ceil(fakepackage)

    def test_divmod(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            divmod(fakepackage, 5)

    def test_floor(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            math.floor(fakepackage)

    def test_floordiv(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage // 2

    def test_iadd(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage += 2

    def test_ifloordiv(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage //= 3

    def test_imod(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage %= 2

    def test_imul(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage *= 2

    def test_ipow(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage **= 5

    def test_isub(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage -= 3

    def test_itruediv(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage /= 2

    def test_mod(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage % 2

    def test_mul(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage * 2

    def test_neg(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            -fakepackage

    def test_pos(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            +fakepackage

    def test_pow(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            math.pow(fakepackage, 2)

    def test_radd(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            object() + fakepackage

    def test_rdiv(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            object() / fakepackage

    def test_rdivmod(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            divmod(object(), fakepackage)

    def test_rfloordiv(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            object() // fakepackage

    def test_rmod(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            object() % fakepackage

    def test_rmul(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            object() * fakepackage

    def test_round(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            round(fakepackage)

    def test_rpow(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            math.pow(5, fakepackage)

    def test_rsub(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            object() - fakepackage

    def test_rtruediv(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            object() / fakepackage

    def test_sub(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage - 5

    def test_truediv(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage / 5

    def test_trunc(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            math.trunc(fakepackage)











import os
import sys
from unittest import skip

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_format(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            format(fakepackage)

        with self.assertRaises(MissingDependencyException):
            "{}".format(fakepackage)

    def test_fspath(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            os.fspath(fakepackage)

    def test_repr(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            repr(fakepackage)

    def test_str(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            str(fakepackage)











import importlib
import pkgutil
import sys
from inspect import getmodule
from pathlib import Path


def get_installed_modules_names():
    """ https://stackoverflow.com/a/73958089/3936044 """
    iter_modules = {module.name for module in pkgutil.iter_modules()}
    builtin = sys.builtin_module_names
    return set.union(iter_modules, builtin)

def module_is_installed(*names):
    """ Returns whether a package is installed.
        `find_spec(name) is None` was the previous solution but namespaces returned True. """
    for name in names:
        spec = get_spec(name)
        if not spec or spec_is_namespace(spec=spec):
            return False
    return True

def _safe_import(name):
    try:
        return importlib.import_module(name=name)
    except (ModuleNotFoundError, TypeError, ImportError) as e:
        return None

def import_module(name, error=True):
    """ Like importlib.import_module with optional error paremeter to return None if errored.
        Also excludes namespaces. """
    module = _safe_import(name=name)
    if module_is_namespace(module=module):
        if error:
            raise ModuleNotFoundError(f"Module '{name}' is only a namespace.")
        return None
    else:
        if module is None and error:
            raise ModuleNotFoundError(f"Module '{name}' isn't installed.")
        return module

def get_spec(fullname):
    return importlib.util.find_spec(fullname)

def spec_is_namespace(spec):
    return spec and spec.loader is None

def module_is_namespace(module):
    """ Returns if given module is a namespace. """
    return hasattr(module, "__path__") and getattr(module, "__file__", None) is None

def module_name_is_namespace(name):
    """ Checks if module's name is a namespace without adding it to sys.modules. """
    was_in_modules = name in sys.modules
    module = _safe_import(name=name)
    is_namespace = module_is_namespace(module=module)

    if was_in_modules:
        sys.modules.pop(name, None)

    return is_namespace

def fake_module_check(obj, error=True):
    """ Simple assertion to trigger error_func earlier if module isn't installed. """
    if type(obj).__name__ == "FakeModule":
        if error:
            obj.error_func("fake_module_check")
        else:
            return True
    else:
        return False


def _get_previous_frame_filename(depth):
    frame = sys._getframe(depth)
    files = ("importlib", "generalimport_bottom.py")

    while frame:
        filename = frame.f_code.co_filename
        frame_is_origin = all(file not in filename for file in files)
        if frame_is_origin:
            return filename
        frame = frame.f_back

def _module_in_stack(module):
    frame = sys._getframe(0)
    while frame:
        if getmodule(frame) is module:
            return True
        frame = frame.f_back
    return False

def _get_scope_from_filename(filename):
    last_part = Path(filename).parts[-1]
    return filename[0:filename.index(last_part)]

def _get_top_name(fullname):
    return fullname.split(".")[0]



from typing import Optional
import logging
from functools import partialmethod

from generalimport.exception import MissingDependencyException
from generalimport.dunders import DynamicDunder, NON_CALLABLE_DUNDERS, CALLABLE_CLASS_DUNDERS, CALLABLE_DUNDERS
from generalimport.import_catcher import ErrorPars

EXCEPTION_NAMING_PATTERNS = ["Exception", "Error"]

logger = logging.getLogger("generalimport")


class FakeModule:
    """ Behaves like a module but any attrs asked for always returns self.
        Raises a ModuleNotFoundError when used in any way.
        Unhandled use-cases: https://github.com/ManderaGeneral/generalimport/issues?q=is%3Aissue+is%3Aopen+label%3Aunhandled """
    __path__ = []
    # __args__ = []  # Doesn't seem necessary
    SENTINEL = object()

    def __init__(self, spec, trigger: Optional[str] = None, catcher=None):
        self.name = spec.name
        self.trigger = trigger
        self.catcher = catcher

        self.__name__ = spec.name
        self.__loader__ = spec.loader
        self.__spec__ = spec
        self.__fake_module__ = True  # Should not be needed, but let's keep it for safety?


    @classmethod
    def _dynamic_dunder_check(cls, error_pars: ErrorPars):
        for dynamic_dunder_cls in DynamicDunder.subclasses:
            dynamic_dunder = dynamic_dunder_cls(error_pars=error_pars)
            if dynamic_dunder.trigger():
                return dynamic_dunder.result()

        return cls.SENTINEL

    @classmethod
    def _get_error_message(cls, error_pars: ErrorPars):
        required_by = f" (required by '{error_pars.trigger}')" if error_pars.trigger else ""
        name_part = f"{error_pars.name}{required_by} " if error_pars.name else ""

        msg_list = [
            f"Optional dependency {name_part}was used but it isn't installed.",
            f"Triggered by '{error_pars.caller}'.",
        ]
        if error_pars.catcher is not None and error_pars.catcher.message:
            msg_list.append(error_pars.catcher.message)

        return " ".join(msg_list)

    @classmethod
    def _error_func(cls, error_pars):
        result = cls._dynamic_dunder_check(error_pars=error_pars)
        if result is not cls.SENTINEL:
            return result

        msg = cls._get_error_message(error_pars=error_pars)

        logger.debug(msg=msg)
        raise MissingDependencyException(msg=msg)

    def error_func(self, _caller: str, *args, **kwargs):
        error_pars = ErrorPars(name=self.name, trigger=self.trigger, caller=_caller, catcher=self.catcher, args=args, kwargs=kwargs)
        return self._error_func(error_pars=error_pars)

    @classmethod
    def error_func_class(cls, _caller: str, *args, **kwargs):
        error_pars = ErrorPars(name=None, trigger=None, caller=_caller, catcher=None, args=args, kwargs=kwargs)
        return cls._error_func(error_pars=error_pars)

    @staticmethod
    def _item_is_exception(item):
        return any(str(item).endswith(pattern) for pattern in EXCEPTION_NAMING_PATTERNS)

    @staticmethod
    def _item_is_dunder(item):
        return item in NON_CALLABLE_DUNDERS

    def __getattr__(self, item):
        fakemodule = FakeModule(spec=self.__spec__, trigger=item, catcher=self.catcher)
        if self._item_is_exception(item=item) or self._item_is_dunder(item=item):
            return fakemodule.error_func(item)
        return fakemodule


# Sets all the callable dunders of FakeModule to 'error_func()' by preserving the name of the dunder that triggered it.
# Mainly useful for debug purposes.
for dunder in CALLABLE_DUNDERS:
    setattr(FakeModule, dunder, partialmethod(FakeModule.error_func, dunder))

for dunder in CALLABLE_CLASS_DUNDERS:
    setattr(FakeModule, dunder, partialmethod(FakeModule.error_func_class, FakeModule, dunder))







import sys
from unittest import skip, IsolatedAsyncioTestCase

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(IsolatedAsyncioTestCase, ImportTestCase):
    async def test_aenter(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            async with fakepackage():
                pass

    async def test_aiter(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            async for x in fakepackage:
                pass

    async def test_await(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            await fakepackage











from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_annotations(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage.__annotations__

    def test_call(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage()

    def test_closure(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage.__closure__

    def test_code(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage.__code__

    def test_defaults(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage.__defaults__

    def test_globals(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage.__globals__

    def test_kwdefaults(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage.__kwdefaults__




import sys
from unittest import skip

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_and(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage & "foo"

    def test_iand(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage &= "foo"

    def test_ior(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage |= "foo"

    def test_or(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage | "foo"

    def test_rand(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            "foo" & fakepackage

    def test_ror(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            "foo" | fakepackage

    def test_rxor(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            "foo" ^ fakepackage

    def test_xor(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage ^ "foo"












from generalimport.generalimport_bottom import get_installed_modules_names, module_is_installed, import_module, module_name_is_namespace, module_is_namespace, spec_is_namespace, get_spec, fake_module_check
from generalimport.exception import MissingDependencyException, MissingOptionalDependency
from generalimport.dunders import DynamicDunder, NON_CALLABLE_DUNDERS, CALLABLE_CLASS_DUNDERS, CALLABLE_DUNDERS
from generalimport.fake_module import FakeModule
from generalimport.general_importer import GeneralImporter
from generalimport.import_catcher import ImportCatcher
from generalimport.top import generalimport, get_importer, reset_generalimport, is_imported



import sys
from unittest import skip

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_class_getitem(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            type(fakepackage)[5]

    def test_dir(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            dir(fakepackage)

    # https://github.com/ManderaGeneral/generalimport/issues/8
    @skip("Cannot cover __setattr__ as it won't allow `import foo.bar`.")
    def test_setattr(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage.x = 5














""" Standalone minimal example """

import sys
import importlib
from unittest.case import SkipTest
from _pytest.outcomes import Skipped

class MissingDependencyException(SkipTest, Skipped):
    def __init__(self, msg=None):
        self.msg = msg
    def __repr__(self):
        return f"MissingDependencyException: {self.msg}" if self.msg else f"MissingDependencyException"

class GeneralImporter:
    def __init__(self, *names):
        self.names = names
        sys.meta_path.insert(0, self)
    def find_spec(self, fullname, path=None, target=None):
        if fullname in self.names:
            return importlib.util.spec_from_loader(fullname, self)
    def create_module(self, spec):
        return FakeModule(name=spec.name)
    def exec_module(self, module):
        pass

class FakeModule:
    def __init__(self, name):
        self.name = name
    def __call__(self, *args, **kwargs):
        raise MissingDependencyException(f"Optional dependency '{self.name}' was used but it isn't installed.")

GeneralImporter("notinstalled")
import notinstalled  # No error
print(notinstalled)  # <__main__.FakeModule object at 0x0000014B7F6D9E80>
notinstalled()  # MissingDependencyException: Optional dependency 'notinstalled' was used but it isn't installed.



import sys

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_len(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            len(fakepackage)

    def test_next(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            next(fakepackage)

    def test_reversed(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            reversed(fakepackage)

    def test_contains(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            5 in fakepackage

    def test_getitem(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage[0]

    def test_index(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage.__index__()

    def test_setitem(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage[5] = "foo"



import sys
from unittest import skip

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_init_subclass(self):
        """ This one is caught by __call__. """
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            class X(fakepackage):
                pass











import dataclasses
from logging import getLogger
from typing import Optional, Union

from generalimport.generalimport_bottom import _get_previous_frame_filename, _get_top_name, _get_scope_from_filename


class ImportCatcher:
    WILDCARD = "*"

    def __init__(self, *names, message=None):
        self.names = set(names)
        self.message = message

        self.added_names = set()
        self.added_fullnames = set()
        self.enabled = True
        self._scope = self._get_scope()

        getLogger(__name__).info(f"Created Catcher with names {self.names} and scope {self._scope}")

        self.latest_scope_filename = None

    def handle(self, fullname):
        if not self._handle_name(fullname=fullname):
            return False
        if not self._handle_scope():
            return False

        self._store_handled_name(fullname=fullname)
        return True



    @staticmethod
    def _get_scope():
        filename = _get_previous_frame_filename(depth=4)
        return _get_scope_from_filename(filename=filename)

    def _store_handled_name(self, fullname):
        name = _get_top_name(fullname=fullname)
        self.added_names.add(name)
        self.added_fullnames.add(fullname)

    def _handle_name(self, fullname):
        name = _get_top_name(fullname=fullname)
        if self.WILDCARD in self.names:
            return True
        if name in self.names:
            return True
        return False

    def _handle_scope(self):
        if self._scope is None:
            return True
        filename = _get_previous_frame_filename(depth=6)
        self.latest_scope_filename = filename
        return filename.startswith(self._scope)


@dataclasses.dataclass
class ErrorPars:
    name: Optional[str]
    trigger: Optional[str]
    caller: str
    catcher: Optional[ImportCatcher]
    args: Union[list, tuple]
    kwargs: dict



import sys
from unittest import skip

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_delattr(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            del fakepackage.foo

    def test_delitem(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            del fakepackage[5]











import sys
from unittest import skip

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_eq(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage == 2

    def test_ge(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage >= 2

    def test_gt(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage > 2

    # https://github.com/ManderaGeneral/generalimport/issues/4
    @skip("Cannot cover `raise X` unless `isinstance(X, Y)` is covered.")
    def test_instancecheck(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            isinstance(fakepackage, int)

        with self.assertRaises(MissingDependencyException):
            raise fakepackage  # This should work if we solve isinstance()

    def test_le(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage <= 2

    def test_lt(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage < 2

    def test_ne(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage != 2

    def test_subclasscheck(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            issubclass(int, fakepackage)



import sys
from typing import Optional
from unittest import skip

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):

    def test_type_with_module(self):
        generalimport("fakepackage")
        import fakepackage

        def my_function1(something: Optional[fakepackage]):
            pass
        
        def my_function2(something: int) -> Optional[fakepackage]:
            pass

        my_function1(1)
        my_function2(1)


    def test_type_with_attribute(self):
        generalimport("fakepackage")
        import fakepackage

        def my_function1(something: Optional[fakepackage.something]):
            pass

        def my_function2(something) -> Optional[fakepackage.something]:
            pass

        my_function1(1)
        my_function2(1)






import sys
from unittest import skip

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_enter(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            with fakepackage:
                pass

        with self.assertRaises(MissingDependencyException):
            fakepackage.__enter__()

    def test_exit(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage.__exit__()











import typing

from generalimport.generalimport_bottom import _module_in_stack
from generalimport.import_catcher import ErrorPars


class DynamicDunder:
    """ Inherit to define a dynamic dunder.
        All subclasses' triggers are tested for truthy before a MissingOptionalDependency is raised.
        Returns result() of first triggered dynamic dunder. """
    subclasses: typing.List[typing.Type["DynamicDunder"]] = []

    def __init_subclass__(cls, **kwargs):
        cls.subclasses.append(cls)

    def __init__(self, error_pars: ErrorPars):
        self.error_pars = error_pars

    def trigger(self): return True
    def result(self): ...


NON_CALLABLE_DUNDERS = (
    # Callable
    "__annotations__", "__closure__", "__code__", "__defaults__", "__globals__", "__kwdefaults__",
    # Info
    "__bases__", "__class__", "__dict__", "__doc__", "__module__", "__name__", "__qualname__", "__all__", "__slots__",
    # Pydantic
    "_nparams",
)
CALLABLE_CLASS_DUNDERS = [
    # Lookup
    "__class_getitem__",
]
CALLABLE_DUNDERS = [
    # Binary
    "__ilshift__", "__invert__", "__irshift__", "__ixor__", "__lshift__", "__rlshift__", "__rrshift__", "__rshift__",
    # Callable
    "__call__",
    # Cast
    "__bool__", "__bytes__", "__complex__", "__float__", "__int__", "__iter__", "__hash__",
    # Compare
    "__eq__", "__ge__", "__gt__", "__instancecheck__", "__le__", "__lt__", "__ne__", "__subclasscheck__",
    # Context
    "__enter__", "__exit__",
    # Delete
    "__delattr__", "__delitem__", "__delslice__",
    # Info
    "__sizeof__", "__subclasses__",
    # Iterable
    "__len__", "__next__", "__reversed__", "__contains__", "__getitem__", "__setitem__",
    # Logic
    "__and__", "__iand__", "__ior__", "__or__", "__rand__", "__ror__", "__rxor__", "__xor__",
    # Lookup
    "__dir__",
    # Math
    "__abs__", "__add__", "__ceil__", "__divmod__", "__floor__", "__floordiv__", "__iadd__", "__ifloordiv__",
    "__imod__", "__imul__", "__ipow__", "__isub__", "__itruediv__", "__mod__", "__mul__", "__neg__", "__pos__",
    "__pow__", "__radd__", "__rdiv__", "__rdivmod__", "__rfloordiv__", "__rmod__", "__rmul__", "__round__",
    "__rpow__", "__rsub__", "__rtruediv__", "__sub__", "__truediv__", "__trunc__",
    # Matrix
    "__imatmul__", "__matmul__", "__rmatmul__",
    # Object
    "__init_subclass__", "__prepare__", "__set_name__",
    # Pickle
    "__getnewargs__", "__getnewargs_ex__", "__getstate__", "__reduce__", "__reduce_ex__",
    # String
    "__format__", "__fspath__", "__repr__", "__str__",
    # Thread
    "__aenter__", "__aexit__", "__aiter__", "__anext__", "__await__",
    # Typing
    "__origin__",
]


class DynamicEQ(DynamicDunder):
    def trigger(self):
        return _module_in_stack(module=typing) and self.error_pars.caller == "__eq__" and bool(self.error_pars.args)

    def result(self):
        other = self.error_pars.args[0]
        return id(self) == id(other)


class DynamicHash(DynamicDunder):
    def trigger(self):
        return _module_in_stack(module=typing) and self.error_pars.caller == "__hash__"

    def result(self):
        return id(self)



import importlib
import sys
from logging import getLogger

from generalimport.fake_module import FakeModule
from generalimport.generalimport_bottom import spec_is_namespace, _get_top_name, get_spec, fake_module_check, module_is_namespace


class GeneralImporter:
    """ Creates fake packages if they don't exist.
        These fake packages' attrs are always a function that raises a ModuleNotFoundError when used.
        This lets you write a single line to handle all your optional dependencies.
        If wildcard (default "*") is provided then this will work on any missing package. """

    singleton_instance = None

    def __init__(self):
        self.catchers = []
        self.latest_catcher = None

        self._singleton()
        self._skip_fullname = None
        sys.meta_path.insert(0, self)

    def catch(self, fullname):
        """ Return first catcher that handles given fullname and filename.

            :rtype: generalimport.ImportCatcher """
        for catcher in self.catchers:
            if catcher.handle(fullname=fullname):
                return catcher

    def find_spec(self, fullname, path=None, target=None):
        if self._ignore_next_import(fullname=fullname):
            return self._handle_ignore(fullname=fullname, reason="Recursive break")

        if self._ignore_existing_top_name(fullname=fullname):
            return self._handle_ignore(fullname=fullname, reason="Top name exists and is not namespace")

        spec = get_spec(fullname)
        if not spec:
            return self._handle_handle(fullname=fullname, reason="Doesn't exist")

        if spec_is_namespace(spec=spec):
            return self._handle_handle(fullname=fullname, reason="Namespace package")

        return self._handle_relay(fullname=fullname, spec=spec)

    def create_module(self, spec):
        return FakeModule(spec=spec, catcher=self.latest_catcher)

    def exec_module(self, module):
        pass



    def _singleton(self):
        assert self.singleton_instance is None
        GeneralImporter.singleton_instance = self

    def _ignore_existing_top_name(self, fullname):
        name = _get_top_name(fullname=fullname)
        if name == fullname:
            return False
        module = sys.modules.get(name, None)
        module_is_real = not fake_module_check(module, error=False)
        return module_is_real and not module_is_namespace(module)

    def _ignore_next_import(self, fullname):
        if fullname == self._skip_fullname:
            self._skip_fullname = None
            return True
        else:
            self._skip_fullname = fullname
            return False

    def _handle_ignore(self, fullname, reason):
        getLogger(__name__).debug(f"Ignoring '{fullname}' - {reason}")
        return None

    def _handle_handle(self, fullname, reason):
        self.latest_catcher = self.catch(fullname=fullname)
        if not self.latest_catcher:
            return self._handle_ignore(fullname=fullname, reason="Unhandled")

        getLogger(__name__).info(f"{self.latest_catcher} is handling '{fullname}' - {reason}")

        sys.modules.pop(fullname, None)  # Remove possible namespace

        return importlib.util.spec_from_loader(fullname, self)

    def _handle_relay(self, fullname, spec):
        getLogger(__name__).debug(f"'{fullname}' exists, returning it's spec '{spec}'")
        return spec



import sys

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_ilshift(self):
        generalimport("fakepackage")
        import fakepackage
        with self.assertRaises(MissingDependencyException):
            fakepackage <<= 5

    def test_irshift(self):
        generalimport("fakepackage")
        import fakepackage
        with self.assertRaises(MissingDependencyException):
            fakepackage >>= 5

    def test_invert(self):
        generalimport("fakepackage")
        import fakepackage
        with self.assertRaises(MissingDependencyException):
            ~fakepackage

    def test_ixor(self):
        generalimport("fakepackage")
        import fakepackage
        with self.assertRaises(MissingDependencyException):
            fakepackage ^= 5

    def test_lshift(self):
        generalimport("fakepackage")
        import fakepackage
        with self.assertRaises(MissingDependencyException):
            fakepackage << 5

    def test_rlshift(self):
        generalimport("fakepackage")
        import fakepackage
        with self.assertRaises(MissingDependencyException):
            5 << fakepackage

    def test_rrshift(self):
        generalimport("fakepackage")
        import fakepackage
        with self.assertRaises(MissingDependencyException):
            5 >> fakepackage

    def test_rshift(self):
        generalimport("fakepackage")
        import fakepackage
        with self.assertRaises(MissingDependencyException):
            fakepackage >> 5





""" Call `generalimport` before importing any optional dependencies. """

from generalimport import generalimport
generalimport("notinstalled")

from notinstalled import missing_func  # No error

missing_func()  # Error occurs here

"""
```
MissingDependencyException: Optional dependency 'notinstalled' was used but it isn't installed.
```

Imports fail when they are **used**, *not* imported. 

This means you don't need to keep checking if the package is installed before importing it.
Simply import your optional package and use it like you would any package and let it fail wherever it fails, with a nice error message. """


import sys
from unittest import skip

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_sizeof(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            sys.getsizeof(fakepackage)

    def test_subclasses(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage.__subclasses__()











import unittest

from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_raise_exception(self):
        generalimport("missing_dep")

        from missing_dep import MyException

        with self.assertRaises(MissingDependencyException):
            raise MyException

    def test_raise_error(self):
        generalimport("missing_dep")

        from missing_dep import MyError

        with self.assertRaises(MissingDependencyException):
            raise MyError

    def test_raise_error_attribute(self):
        generalimport("missing_dep")

        import missing_dep

        with self.assertRaises(MissingDependencyException):
            raise missing_dep.MyError

    def test_try_catch(self):
        from generalimport import generalimport
        generalimport("missing_dep")

        from missing_dep import LangDetectException, detect

        try:
            detect("foo")
        except LangDetectException:
            pass

    @unittest.skip("Catching raise currently relies on suffix.")
    def test_raise_error_bad_suffix(self):
        generalimport("missing_dep")

        from missing_dep import ErrorBadSuffix

        with self.assertRaises(MissingDependencyException):
            raise ErrorBadSuffix


""" The beauty of this package is that the error raised isn't just any exception.
    It has two base classes: `unittest.case.SkipTest` and `_pytest.outcomes.Skipped` (If available).

    This means that if a test method uses an uninstalled optional package then that test is automatically skipped.
    This means no more manual skip decorators for optional dependencies! """

from generalimport import generalimport
generalimport("optional_uninstalled_package")

from optional_uninstalled_package import missing_func

from unittest import TestCase

class MyTest(TestCase):
    def test_missing_func(self):
        self.assertEqual(3, missing_func(1, 2))

"""
```
Ran 1 test in 0.002s

OK (skipped=1)

Skipped: Optional dependency 'optional_uninstalled_package' was used but it isn't installed.
```
"""



import importlib
import sys

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase
from generalimport.generalimport_bottom import _safe_import


class Test(ImportTestCase):
    def test_get_installed_packages(self):
        self.assertIn("generalimport", get_installed_modules_names())
        self.assertIn("mmap", get_installed_modules_names())  # built-in
        self.assertNotIn("doesntexist", get_installed_modules_names())

    def test_package_is_installed(self):
        self.assertEqual(True, module_is_installed("generalimport"))
        self.assertEqual(True, module_is_installed("setuptools"))
        self.assertEqual(False, module_is_installed("doesntexist"))
    
    def test_package_is_imported(self):
        self.assertEqual(True, is_imported("generalimport"))
        self.assertEqual(False, is_imported("setuptools"))
        self.assertEqual(False, is_imported("doesntexist"))

    def test_MissingDependencyException(self):
        self.assertEqual("foo", MissingDependencyException("foo").msg)
        self.assertIn("foo", repr(MissingDependencyException("foo")))
        self.assertIn("foo", str(MissingDependencyException("foo")))

        self.assertIn("MissingDependencyException", repr(MissingDependencyException()))
        self.assertEqual(str(AttributeError()), str(MissingDependencyException()))

    def test_GeneralImporter(self):
        generalimport("packagethatdoesntexist")

        self.assertIn(get_importer(), sys.meta_path)

        self.assertRaises(AssertionError, GeneralImporter)  # Singleton

        import packagethatdoesntexist
        self.assertRaises(MissingDependencyException, packagethatdoesntexist)

        from packagethatdoesntexist.module import fakefunc
        self.assertRaises(MissingDependencyException, fakefunc)

        import packagethatdoesntexist.module.fakefunc2
        self.assertRaises(MissingDependencyException, packagethatdoesntexist.module.fakefunc2)

        with self.assertRaises(ImportError):
            import notexisting

        with self.assertRaises(ImportError):
            from notexisting import hi

        with self.assertRaises(ImportError):
            from notexisting.hey import hi

        with self.assertRaises(ImportError):
            import notexisting.hi.hey

    def test_find_module_relay(self):
        """ pkg_resources is one of the few packages always included in venv. """
        generalimport("pkg_resources")
        sys.modules.pop("pkg_resources", None)
        import pkg_resources
        fake_module_check(pkg_resources)

    def test_wildcard(self):
        generalimport("*")

        from whateverz import this
        self.assertRaises(MissingDependencyException, this)

        from whateverz.hi import that
        self.assertRaises(MissingDependencyException, that)

        import hiii
        self.assertRaises(MissingDependencyException, hiii)

        import heyyyy.foo
        self.assertRaises(MissingDependencyException, heyyyy.foo)

    def test_import_module(self):
        self.assertRaises(ModuleNotFoundError, import_module, "thisdoesntexist")

        generalimport("thisdoesntexist")
        thisdoesntexist = import_module("thisdoesntexist")

        self.assertRaises(MissingDependencyException, thisdoesntexist)

        self.assertIs(import_module("generalimport"), gi)
        import_module("generalimport")

    def test_import_module_namespace(self):
        self.assertIs(None, import_module("namespace", error=False))
        self.assertRaises(ModuleNotFoundError, import_module, "namespace")

    def test_namespace_importer(self):
        self.assertEqual(True, module_name_is_namespace("namespace"))
        self.assertEqual(True, module_name_is_namespace("namespace"))

        generalimport("namespace")

        from namespace.mod import func

        self.assertRaises(MissingDependencyException, func)

    def test_generalimport(self):
        self.assertEqual(True, module_name_is_namespace("namespace"))
        self.assertEqual(True, module_name_is_namespace("namespace"))

        catcher = generalimport("namespace", "missing_dep")
        self.assertEqual(2, len(catcher.names))

        catcher2 = generalimport("another_missing")
        self.assertEqual(1, len(catcher2.names))

        import namespace
        import missing_dep
        import another_missing

        self.assertRaises(MissingDependencyException, namespace.func)
        self.assertRaises(MissingDependencyException, missing_dep.func)
        self.assertRaises(MissingDependencyException, another_missing.func)

    def test_fake_module_check(self):
        generalimport("fakepackage")
        import fakepackage

        self.assertRaises(MissingDependencyException, fake_module_check, fakepackage)
        self.assertIs(True, fake_module_check(fakepackage, error=False))

        fake_module_check(sys)
        self.assertIs(False, fake_module_check(sys, error=False))

    def test_import_catcher_scope(self):
        catcher = generalimport("hi")
        self.assertIn("test", catcher._scope)

    def test_import_catcher_outside_scope(self):
        catcher = generalimport("fakepackage")
        catcher._scope = "some/random/path"

        self.assertRaises(ImportError, import_module, "fakepackage")

    def test_import_catcher_scope_is_none(self):
        catcher = generalimport("fakepackage")
        catcher._scope = None

        import fakepackage

        self.assertIs(True, fake_module_check(fakepackage, error=False))

    def test_latest_scope_filename_import(self):
        catcher = generalimport("hi")

        import hi
        self.assertIs(True, fake_module_check(hi, error=False))
        self.assertIn("test_generalimport.py", catcher.latest_scope_filename)

    def test_latest_scope_filename_import_module(self):
        catcher = generalimport("hi")

        hi = import_module("hi")
        self.assertIs(True, fake_module_check(hi, error=False))
        self.assertIn("test_generalimport.py", catcher.latest_scope_filename)

    def test_latest_scope_filename_safe_import(self):
        catcher = generalimport("hi")

        hi = _safe_import("hi")
        self.assertIs(True, fake_module_check(hi, error=False))
        self.assertIn("test_generalimport.py", catcher.latest_scope_filename)

    def test_latest_scope_filename_importlib(self):
        catcher = generalimport("hi")

        hi = importlib.import_module("hi")
        self.assertIs(True, fake_module_check(hi, error=False))
        self.assertIn("test_generalimport.py", catcher.latest_scope_filename)

    def test_correct_message(self):
        from generalimport import generalimport
        generalimport("missing_dep")

    def test_logging_message(self):
        generalimport("nonexisting", "missing_dep")
        import nonexisting

        with self.assertLogs('generalimport', level='DEBUG') as cm:
            with self.assertRaises(MissingDependencyException):
                nonexisting.func()

        self.assertIn("nonexisting", cm.output[0])
        self.assertIn("__call__", cm.output[0])

        from missing_dep import foo
        from missing_dep import bar

        try:
            foo()
        except MissingDependencyException as e:
            self.assertIn("foo", str(e))
        else:
            self.fail("Error not raised")

        try:
            bar()
        except MissingDependencyException as e:
            self.assertIn("bar", str(e))
        else:
            self.fail("Error not raised")

    def test_message(self):
        generalimport("nonexisting", message="FooBar")
        import nonexisting

        try:
            nonexisting()
        except MissingDependencyException as e:
            self.assertIn("FooBar", str(e))
        else:
            self.fail("Error not raised")

    def test_catcher_relay_single(self):
        catcher = generalimport("nonexisting")
        import nonexisting

        self.assertIs(nonexisting.catcher, catcher)

    def test_catcher_relay_two(self):
        catcher = generalimport("foo", "bar")
        import foo
        import bar

        self.assertIs(foo.catcher, catcher)
        self.assertIs(bar.catcher, catcher)

    def test_catcher_relay_double(self):
        catcher = generalimport("foo")
        from foo import bar

        self.assertIs(bar.catcher, catcher)

    def test_catcher_relay_double_and_two(self):
        foobar_catcher = generalimport("foo")
        hiii_catcher = generalimport("hiii")

        from foo import bar
        import hiii

        self.assertIs(bar.catcher, foobar_catcher)
        self.assertIs(hiii.catcher, hiii_catcher)
























import sys
from unittest import skip

import generalimport as gi
from generalimport import *

from generalimport.test.funcs import ImportTestCase


class Test(ImportTestCase):
    def test_imatmul(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage @= 5

    def test_matmul(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            fakepackage @ 2

    def test_rmatmul(self):
        generalimport("fakepackage")
        import fakepackage

        with self.assertRaises(MissingDependencyException):
            2 @ fakepackage












import warnings


def _get_skip_base_classes():
    from unittest.case import SkipTest
    yield SkipTest

    try:
        from _pytest.outcomes import Skipped
        yield Skipped
    except ImportError:
        pass


class SkipTestException(*_get_skip_base_classes()):
    def __init__(self, msg=None):
        if msg is None:
            msg = ""
        self.msg = msg

    def __repr__(self):
        message = f"'{self.msg}'" if self.msg else ""
        return f"{type(self).__name__}({message})"

    def __str__(self):
        return self.msg


class MissingDependencyException(SkipTestException):
    pass


class MissingOptionalDependency(SkipTestException):
    """ MissingOptionalDependency is deprecated, use MissingDependencyException """
    pass

# def MissingOptionalDependency(*args, **kwargs):
#     warnings.warn("MissingOptionalDependency has been changed to MissingDependencyException", DeprecationWarning, stacklevel=2)
#     return MissingDependencyException(*args, **kwargs)





















import sys

from generalimport import FakeModule, MissingDependencyException
from generalimport.general_importer import GeneralImporter
from generalimport.import_catcher import ImportCatcher


def _assert_no_dots(names):
    for name in names:
        assert "." not in name, f"Dot found in '{name}', only provide package without dots."

def get_importer():
    """ Return existing or new GeneralImporter instance. """
    return GeneralImporter.singleton_instance or GeneralImporter()

def generalimport(*names, message=None):
    """ Adds names to a new ImportCatcher instance.
        Creates GeneralImporter instance if it doesn't exist. """
    # print(get_previous_frame_filename())
    _assert_no_dots(names=names)
    catcher = ImportCatcher(*names, message=message)
    get_importer().catchers.append(catcher)
    return catcher



def _pop_imported_modules():
    for catcher in get_importer().catchers:
        for fullname in catcher.added_fullnames:
            sys.modules.pop(fullname, None)

def _clear_importer():
    importer = get_importer()
    sys.meta_path.remove(importer)
    importer.catchers.clear()
    GeneralImporter.singleton_instance = None

def reset_generalimport():
    _pop_imported_modules()
    _clear_importer()


def is_imported(module_name: str) -> bool:
    """
    Returns True if the module was actually imported, False, if generalimport mocked it.
    """
    module = sys.modules.get(module_name)
    try:
        return bool(module and not isinstance(module, FakeModule))
    except MissingDependencyException as exc:
        # isinstance() raises MissingDependencyException: fake module
        pass
    return False
